--[[Author:		Cyprias 
	License:	All Rights Reserved
	Contact:	Cyprias on Curse.com or WowAce.com

	Only Curse.com, Wowace.com and WoWInterface.com have permission to host this addon.
	I have not given permission for Plate Buffs to be used in any addon compilation or UI pack.]]
local folder, core = ...
	
-- global lookup
local _G = _G
local pairs = pairs
local InterfaceOptionsFrame_OpenToCategory = InterfaceOptionsFrame_OpenToCategory
local GetTime = GetTime
local table_sort = table.sort
local string_sub = string.sub
local string_len = string.len
local LibStub = LibStub
local table_getn = table.getn
local UnitGUID = UnitGUID
local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff
local table_insert = table.insert
local UnitExists = UnitExists
local UnitIsUnit = UnitIsUnit
local Debug = core.Debug
local UnitIsPlayer = UnitIsPlayer
local UnitClassification = UnitClassification
local UnitName = UnitName
local date = date
local table_remove = table.remove
local GetUnitName = GetUnitName
local _ --underscore so GetGlobals doesn't nag me.

--~ PlateBuffs = core --global table

local LibNameplate = LibStub("LibNameplate-1.0", true)
if not LibNameplate then	error(folder .. " requires LibNameplate-1.0.") return end

-- local
core.title		= "Plate Buffs"
core.version	= GetAddOnMetadata(folder, "X-Curse-Packaged-Version") or ""
core.titleFull	= core.title.." "..core.version
core.addonDir = "Interface\\AddOns\\PlateBuffs\\"

core.LibNameplate = LibNameplate

LibStub("AceAddon-3.0"):NewAddon(core, folder, "AceConsole-3.0", "AceEvent-3.0") 
--~ L = LibStub("AceLocale-3.0"):GetLocale(folder, true)

core.L = LibStub("AceLocale-3.0"):GetLocale(folder, true)
local L = core.L



local regEvents = {
	"PLAYER_TARGET_CHANGED",
	"UPDATE_MOUSEOVER_UNIT",
	"UNIT_AURA",
	"UNIT_TARGET",
}
core.db	= {}
local db
local P --db.profile
local spellInfo

core.defaultSettings	= {
	profile = {
	},
	global = {
		spellInfo = {},
	},
}

core.buffFrames = {}
core.guidBuffs = {}
core.nametoGUIDs = {}
core.buffBars = {}

local buffBars = core.buffBars
local guidBuffs = core.guidBuffs
local nametoGUIDs = core.nametoGUIDs
local buffFrames = core.buffFrames
local defaultSettings = core.defaultSettings

core.iconTestMode = false
local coreOpts
local spellUI
local profileUI
local whoUI
local barUI

local totemList = {
    2484,--Earthbind Totem
    8143,--Tremor Totem
    8177,--Grounding Totem
    8512,--Windfury Totem
    6495,--Sentry Totem
    8170,--Cleansing Totem
    3738,--Wrath of Air Totem
    2062,--Earth Elemental Totem
    2894,--Fire Elemental Totem
    58734,--Magma Totem
    58582,--Stoneclaw Totem
    58753,--Stoneskin Totem
    58739,--Fire Resistance Totem
    58656,--Flametongue Totem
    58745,--Frost Resistance Totem
    58757,--Healing Stream Totem
    58774,--Mana Spring Totem
    58749,--Nature Resistance Totem
    58704,--Searing Totem
    58643,--Strength of Earth Totem
    57722,--Totem of Wrath
}

local totems = {}
local name, texture
for i=1,table_getn(totemList) do
    name, _, texture = GetSpellInfo(totemList[i])
	totems[i] = {name = name, texture = texture}
end

function core:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("PB_DB", core.defaultSettings)
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileDeleted", "OnProfileChanged")
	self:RegisterChatCommand("pb", "MySlashProcessorFunc")
	
	
	local config = LibStub("AceConfig-3.0")
	local dialog = LibStub("AceConfigDialog-3.0")
	config:RegisterOptionsTable(self.title, self.CoreOptionsTable)
	coreOpts = dialog:AddToBlizOptions(self.title, self.titleFull)
	
	
	
	config:RegisterOptionsTable(self.title.."Who", self.WhoOptionsTable)
	whoUI = dialog:AddToBlizOptions(self.title.."Who", L["Who"], self.titleFull) --Need locale
	
	
	config:RegisterOptionsTable(self.title.."Spells", self.SpellOptionsTable)
	spellUI = dialog:AddToBlizOptions(self.title.."Spells", L["Spells"], self.titleFull) --Need locale

	config:RegisterOptionsTable(self.title.."Bars", self.BarOptionsTable)
	barUI = dialog:AddToBlizOptions(self.title.."Bars", L["Bars"], self.titleFull) --Need locale
	
	
	--last UI
	config:RegisterOptionsTable(self.title.."Profile", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db))
	profileUI = dialog:AddToBlizOptions(self.title.."Profile", L["Profiles"], self.titleFull) --Need locale

end

--[===[@debug@
function PB_NewPlates()--	/run PB_NewPlates()
	for plate, name in pairs(LibNameplate.nameplates) do 
		if plate:IsShown() then
			Debug("OnEnable", plate, name, LibNameplate:GetName(plate))
			core:LibNameplate_NewNameplate("LibNameplate_NewNameplate", plate)
		end
	end
end
--@end-debug@]===]

function core:OnEnable()
	db = self.db
	P = db.profile
	spellInfo = db.global.spellInfo
	
	
	for i, event in pairs(regEvents) do 
		self:RegisterEvent(event)
	end
	
	LibNameplate.RegisterCallback(self, "LibNameplate_NewNameplate")
	LibNameplate.RegisterCallback(self, "LibNameplate_FoundGUID")
	LibNameplate.RegisterCallback(self, "LibNameplate_RecycleNameplate")
	LibNameplate.RegisterCallback(self, "LibNameplate_VisiblePlateChange")
	--Update old options.
	if P.cooldownSize < 6 then
		P.cooldownSize = core.defaultSettings.profile.cooldownSize
	end
	if P.stackSize < 6 then
		P.stackSize = core.defaultSettings.profile.stackSize
	end

--~ 	PB_NewPlates()
	
	for plate in pairs(core.buffBars) do 
		for i=1, table_getn(core.buffBars[plate]) do 
			core.buffBars[plate][i]:Show() --reshow incase user disabled addon.
		end
	end
	

--~ 	Debug("OnEnable", "core", self.db.profile)
end

function core:OnDisable()
	LibNameplate.UnregisterAllCallbacks(self) 
	
	
	for plate in pairs(core.buffBars) do 
		for i=1, table_getn(core.buffBars[plate]) do 
			core.buffBars[plate][i]:Hide() --makesure all frames stop OnUpdating.
		end
	end
end

----------------------------------------------------------------------
function core:OnProfileChanged(...)									--
-- User has reset proflie, so we reset our spell exists options.	--
----------------------------------------------------------------------
	-- Shut down anything left from previous settings
	self:Disable()
	-- Enable again with the new settings
	self:Enable()
end

----------------------------------------------
function core:MySlashProcessorFunc(input)	--
-- /da function brings up the UI options.	--
----------------------------------------------
	InterfaceOptionsFrame_OpenToCategory(spellUI)--open subtab
	InterfaceOptionsFrame_OpenToCategory(coreOpts)
end


function core:LibNameplate_VisiblePlateChange(event, plate, visiblePlate)
--~ 	Debug("VisiblePlateChange", plate, visiblePlate)
--~ 	if buffFrames[plate] then
--~ 		if buffFrames[plate][1] then
--~ 			self:ResetBarPoint(buffFrames[plate][1], plate, visiblePlate)
--~ 		end
--~ 		for i=2, table_getn(buffFrames[plate]) do 
--~ 			buffFrames[plate][i]:SetParent(visiblePlate)
--~ 		end
--~ 	end
	
	if buffBars[plate] then
		if buffBars[plate][1] then
			self:ResetBarPoint(buffBars[plate][1], plate, visiblePlate)
		end
		for i=2, table_getn(buffBars[plate]) do 
			buffBars[plate][i]:SetParent(visiblePlate)
		end
	end
	
end




function core:LibNameplate_RecycleNameplate(event, plate)
	--note to self, not buffBars
	if buffFrames[plate] then
		for i=1, table_getn(buffFrames[plate]) do 
			buffFrames[plate][i]:Hide()
		end
	end
--~ 	local GUID = LibNameplate:GetGUID(plate)
--~ 	if GUID and guidBuffs[GUID] then
--~ 		for i=table.getn(guidBuffs[GUID]), 1, -1 do 
--~ 			if guidBuffs[GUID][i].name == "UNKNOWN" then
--~ 				table.remove(guidBuffs[GUID], i)
--~ 				Debug("Removing icon...")
--~ 			end
--~ 		end
--~ 	end
--~ 	Debug("RecycleNameplate", LibNameplate:GetName(plate))
end

local function isTotem(name)
    for i=1,table_getn(totems) do
        if name:find(totems[i].name) then
            return totems[i].texture
        end
    end
    
    return nil
end

function core:ShouldAddBuffs(plate)
	local plateName =  LibNameplate:GetName(plate) or "UNKNOWN"
--~ 	Debug("ShouldAddBuffs 1", plateName)
	
	if P.showTotems == false and isTotem(plateName) then
--~ 		Debug("ShouldAddBuffs", plateName, "totem")
		return false
	end
	local plateType = LibNameplate:GetType(plate)
	local plateReaction = LibNameplate:GetReaction(plate)

--~ 	Debug("ShouldAddBuffs 2", plateName, plateType, plateReaction)
	
	if (P.abovePlayers == true and plateType == "PLAYER") or (P.aboveNPC == true and plateType == "NPC") then
		if P.aboveFriendly == true and plateReaction == "FRIENDLY" then
			return true
		elseif P.aboveNeutral == true and plateReaction == "NEUTRAL" then
			return true
		elseif P.aboveHostile == true and plateReaction == "HOSTILE" then
			return true
		end
	end
	
	return false
end
	
function core:LibNameplate_NewNameplate(event, plate)
--~ 	Debug("NewNameplate 1", plate)
	if self:ShouldAddBuffs(plate) == true then
		local plateName =  LibNameplate:GetName(plate) or "UNKNOWN"
--~ 		Debug("NewNameplate 2", plate, plateName)
		
		if P.saveNameToGUID == true and nametoGUIDs[plateName] and (LibNameplate:GetType(plate) == "PLAYER" or LibNameplate:IsBoss(plate)) then
			self:AddBuffsToPlate(plate, nametoGUIDs[plateName])
		elseif P.unknownSpellDataIcon == true then
			self:AddUnknownIcon(plate)
		end
	end

end

function core:LibNameplate_FoundGUID(event, plate, GUID, unitID)
	if self:ShouldAddBuffs(plate) == true then
--~ 		local plateName =  LibNameplate:GetName(plate) or "UNKNOWN"
--~ 		Debug("FoundGUID", plateName, plate)
		self:AddBuffsToPlate(plate, GUID)
	end

	
	--[===[@debug@
--~ 	local plateName =  LibNameplate:GetName(plate)
--~ 	Debug("FoundGUID", plateName, plate:GetFrameLevel(), plate:GetFrameStrata())
--~ 	Debug("FoundGUID", LibNameplate:GetName(plate))

--~ 	Debug("FoundGUID", plateName, 
--~ 		GUID, 
--~ 		unitID,
--~ 		LibNameplate:GetLevel(plate),
--~ 		LibNameplate:GetReaction(plate),
--~ 		LibNameplate:GetType(plate),
--~ 		LibNameplate:IsBoss(plate),
--~ 		LibNameplate:GetHealthMax(plate),
--~ 		LibNameplate:GetClass(plate),
--~ 		LibNameplate:IsElite(plate),
--~ 		LibNameplate:GetThreatSituation(plate),
--~ 		LibNameplate:GetAlpha(plate),
--~ 		LibNameplate:IsTarget(plate),
--~ 		LibNameplate:GetHealth(plate), 
--~ 		LibNameplate:GetHealthMax(plate),
--~ 		LibNameplate:IsMouseover(plate),
--~ 		LibNameplate:IsCasting(plate),
--~ 		LibNameplate:IsInCombat(plate),
--~ 		LibNameplate:IsMarked(plate),
--~ 		LibNameplate:GetRaidIcon(plate),
--~ 		LibNameplate:GetGUID(plate)
--~ 		"" --nada
--~ 	)

	--Do nothing values to test library.
	local value = LibNameplate:GetName(plate)
	value = LibNameplate:GetLevel(plate)
	value = LibNameplate:GetReaction(plate)
	value = LibNameplate:GetType(plate)
	value = LibNameplate:IsBoss(plate)
	value = LibNameplate:GetHealthMax(plate)
	value = LibNameplate:GetClass(plate)
	value = LibNameplate:IsElite(plate)
	value = LibNameplate:GetThreatSituation(plate)
	value = LibNameplate:GetAlpha(plate)
	value = LibNameplate:IsTarget(plate)
	value = LibNameplate:GetHealth(plate)
	value = LibNameplate:GetHealthMax(plate)
	value = LibNameplate:IsMouseover(plate)
	value = LibNameplate:IsCasting(plate)
	value = LibNameplate:IsInCombat(plate)
	value = LibNameplate:IsMarked(plate)
	value = LibNameplate:GetRaidIcon(plate)
	value = LibNameplate:GetGUID(plate)
	
	
--~ 	value = 

--~ 	Debug("LibNameplate_FoundGUID")
--~ 	Debug("GetName", LibNameplate:GetName(plate))
--~ 	Debug("GetLevel", LibNameplate:GetLevel(plate))
--~ 	Debug("GetReaction", LibNameplate:GetReaction(plate))
--~ 	Debug("GetType", LibNameplate:GetType(plate))
--~ 	Debug("IsBoss", LibNameplate:IsBoss(plate))
--~ 	Debug("GetHealthMax", LibNameplate:GetHealthMax(plate))
--~ 	Debug("GetClass", LibNameplate:GetClass(plate))
--~ 	Debug("IsElite", LibNameplate:IsElite(plate))
--~ 	Debug("GetThreatSituation", LibNameplate:GetThreatSituation(plate))
--~ 	Debug("GetAlpha", LibNameplate:GetAlpha(plate))
--~ 	Debug("IsTarget", LibNameplate:IsTarget(plate))
--~ 	Debug("GetHealth", LibNameplate:GetHealth(plate), "GetHealthMax", LibNameplate:GetHealthMax(plate))
--~ 	Debug("IsMouseover", LibNameplate:IsMouseover(plate))
--~ 	Debug("IsCasting", LibNameplate:IsCasting(plate))
--~ 	Debug("IsInCombat", LibNameplate:IsInCombat(plate))
--~ 	Debug("IsMarked", LibNameplate:IsMarked(plate))
--~ 	Debug("GetRaidIcon", LibNameplate:GetRaidIcon(plate))
--~ 	Debug("GetGUID", LibNameplate:GetGUID(plate))
	--@end-debug@]===]
	
end

--[===[@debug@
function PB_TargetTest()--	/run PB_TargetTest()
	local unitID = "target"
	local GUID = UnitGUID(unitID)
		local name = UnitName(unitID)
	local plate = LibNameplate:GetNameplateByGUID(GUID)
	
	Debug(GUID, plate, name, plate and LibNameplate:GetName(plate))
	
end

function PB_ResetSpelLInfo()--	/run PB_ResetSpelLInfo()
	spellInfo = {}
end
--@end-debug@]===]

--P.spellOpts
local function CollectUnitInfo(unitID)
	local GUID = UnitGUID(unitID)

--~ 	core:Debug("CollectUnitInfo", unitID, GUID)


	guidBuffs[GUID] = guidBuffs[GUID] or {}

	--Remove all the entries this way instead of replacing the table. I'm trying to lower my garbage.
	
	--[[	]]
	for i=table_getn(guidBuffs[GUID]), 1, -1 do 
		table_remove(guidBuffs[GUID], i)
	end

	local i = 1;
	local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId
	while UnitBuff(unitID, i) do 
		name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId = UnitBuff(unitID, i)
		icon = icon:gsub("Interface\\Icons\\", "")
		
		if P.watchCombatlog == true and not spellInfo[spellId] then
			spellInfo[spellId] = {}
			spellInfo[spellId].icon		= icon
			if duration > 0 then
				spellInfo[spellId].duration	= duration
			end
		end
		
		if P.spellOpts[name] and P.spellOpts[name].show then
			if P.spellOpts[name].show == 1 or (P.spellOpts[name].show == 2 and unitCaster and unitCaster == "player") then
				table_insert(guidBuffs[GUID], {
					name = name, --needed for spell options.
					icon = icon,
					expirationTime = expirationTime,
					startTime = expirationTime - duration,
					duration = duration,
					playerCast = unitCaster and unitCaster == "player" and 1,
					stackCount = count,
					sID = spellId,
					caster = unitCaster and GetUnitName(unitCaster),
				})
			end
		else
			if P.defaultBuffShow == 1 or (P.defaultBuffShow == 2 and unitCaster and unitCaster == "player") then
				table_insert(guidBuffs[GUID], {
					name = name,
					icon = icon,
					expirationTime = expirationTime,
					startTime = expirationTime - duration,
					duration = duration,
					playerCast = unitCaster and unitCaster == "player" and 1,
					stackCount = count,
					sID = spellId,
					caster = unitCaster and GetUnitName(unitCaster),
				})
			end
		end
		
		
		i=i + 1
	end
	
	
	i = 1;
	local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId
	while UnitDebuff(unitID, i) do 
		name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId = UnitDebuff(unitID, i) 
		icon = icon:gsub("Interface\\Icons\\", "")
		
		if P.watchCombatlog == true and not spellInfo[spellId] then
			spellInfo[spellId] = {}
			spellInfo[spellId].icon		= icon
			if duration > 0 then --auras like mounts don't have duration.
				spellInfo[spellId].duration	= duration
			end
				--I tried using spellSchool in combatlog to get debuff type, but spells with same school can have same debuff type.
			spellInfo[spellId].debuffType = debuffType
		end
		

		if P.spellOpts[name] and P.spellOpts[name].show then
			if P.spellOpts[name].show == 1 or (P.spellOpts[name].show == 2 and unitCaster and unitCaster == "player") then
				table_insert(guidBuffs[GUID], {
					name = name, --needed for spell options.
					icon = icon,
					expirationTime = expirationTime,
					startTime = expirationTime - duration,
					duration = duration,
					playerCast = unitCaster and unitCaster == "player" and 1,
					stackCount = count,
					debuffType = debuffType,
					isDebuff	= true,
					sID = spellId,
					caster = unitCaster and GetUnitName(unitCaster),
				})
			end
		else
			if P.defaultDebuffShow == 1 or (P.defaultDebuffShow == 2 and unitCaster and unitCaster == "player") then
				table_insert(guidBuffs[GUID], {
					name = name, --needed for spell options.
					icon = icon,
					expirationTime = expirationTime,
					startTime = expirationTime - duration,
					duration = duration,
					playerCast = unitCaster and unitCaster == "player" and 1,
					stackCount = count,
					debuffType = debuffType,
					isDebuff	= true,
					sID = spellId,
					caster = unitCaster and GetUnitName(unitCaster),
				})
			end
		end
		i=i + 1
	end

	
	
	--[===[@debug@
--~ 	if unitID == "target" then
--~ 		for i=1, table_getn(guidBuffs[GUID]) do 
--~ 			Debug(guidBuffs[GUID][i].name)
--~ 		end
--~ 	end
	--@end-debug@]===]
	

	
	if core.iconTestMode == true then
		for i=table_getn(guidBuffs[GUID]), 1, -1 do 
			for t=1, P.iconsPerBar - 1 do 
				table_insert(guidBuffs[GUID], i, guidBuffs[GUID][i]) --reinsert the entry abunch of times.
			end
		end

	end

	if P.saveNameToGUID == true and UnitIsPlayer(unitID) or UnitClassification(unitID) == "worldboss" then
		--These should always have unique names.
		nametoGUIDs[UnitName(unitID)] = GUID
	end

	
--~ 	local plate = LibNameplate:GetNameplateByGUID(GUID)
--~ 	if plate and core:ShouldAddBuffs(plate) == true then
--~ 		core:AddBuffsToPlate(plate, GUID)
--~ 	end
	core:UpdatePlateByGUID(GUID)
	
end



function core:PLAYER_TARGET_CHANGED(event, ...)
	if UnitExists("target") then
		CollectUnitInfo("target")
	end
end


function core:UNIT_TARGET(event, unitID)--this fires half a second after PLAYER_TARGET_CHANGED. Not good for rapid targeting.
	if not UnitIsUnit(unitID, "player") and UnitExists(unitID.."target") then
		CollectUnitInfo(unitID.."target")
	end
--~ 	Debug(event, ...)
end


function core:UPDATE_MOUSEOVER_UNIT(event, ...)
	if UnitExists("mouseover") then
		CollectUnitInfo("mouseover")
	end
--~ 	Debug(event, ...)
end

function core:UNIT_AURA(event, unitID)
	if UnitExists(unitID) then
		CollectUnitInfo(unitID)
	end
--~ 	Debug(event, ...)
end



function core:AddNewSpell(spellName)
	Debug("AddNewSpell", spellName)
	
	if not P.spellOpts[spellName] then
		P.spellOpts[spellName] = {
--~ 			show = P.defaultSpells,
--~ 			iconSize = P.iconSize,
--~ 			cooldownSize = P.cooldownSize,
--~ 			stackSize = P.stackSize,

			show = 1, --always show
			when = date("%c"),
		}
	else
		self.echo(spellName.." is already in the list.")
	end

	
	self:BuildSpellUI()
end

function core:RemoveSpell(spellName)
	P.spellOpts[spellName] = nil
	core:BuildSpellUI()
end

function core:UpdatePlateByGUID(GUID)
	local plate = LibNameplate:GetNameplateByGUID(GUID)
	if plate and self:ShouldAddBuffs(plate) == true then
		core:AddBuffsToPlate(plate, GUID)
	end
end